# coding: utf-8
from __future__ import unicode_literals
from ..gotomagic.text import GotoError, GotoWarning

def help():
    return "{0:10}{1:30}{2}".format('rm', '<magicword>', 'Remove shortcut')


def names():
    return ['rm','--rm', 'remove','--remove']


def run(magic, command, args, options):
    """
    Remove magicword
    """

    if (len(args) == 0):
        return None, GotoWarning("missing_magicword", command='rm')

    word = args[0]

    err = magic.remove_shortcut(word)
    if err:
        return None, err

    # TODO: move save inside magic.
    err = magic.save()
    if err:
        return None, err
    return 'Removed magic word %s' % word, None



from functools import reduce

from .commands import commands_list
from ..plugins import plugins_list


def help():
    return "{0:10}{1}".format('help', 'Show this message again')


def names():
    return ['help', '--help', '-h', '/?']


def run(magic, command, args, options):
    """
    Get information about usage
    """
    header = """
Goto - the magic traveler, how may I help you?

Wondering how to change project?
    project help                  Consult my brother in command

Basic usage
    goto <magicword>      Go to shortcut
    goto [<magicword>...] Go to many shortcuts
"""

    commands_help = map(lambda x: x.help(), commands_list)
    commands_sorted = sorted(commands_help, key=lambda x: x.startswith('-'), reverse=False)
    commands_text = reduce(lambda x,y: x + "{0:>8} {1}\n".format('goto', y), commands_sorted, '\nCommands\n')

    plugins_help = map(lambda x: x.help(), plugins_list)
    plugins_sorted = sorted(plugins_help, key=lambda x: x.startswith('-'), reverse=False)
    plugins_text = reduce(lambda x,y: x + "{0:>8} {1}\n".format('goto', y), plugins_sorted, '\nPlugins\n')

    return "{}{}{}".format(header, commands_text, plugins_text), None



#  coding: utf-8
"""
Magic is stored here.
"""
from __future__ import absolute_import, unicode_literals
from builtins import dict, str  # redefine dict and str to be py3-like in py2.
# http://johnbachman.net/building-a-python-23-compatible-unicode-sandwich.html

import json
import codecs
import os
import sys

from .text import GotoWarning, GotoError
from .. import settings
from . import utils


class GotoMagic():
    """ Goto's Black box.

        Nah, not really, it is just the place where
        magic shortcuts are created, updated, removed and
        used.

        tl-dr: saving magic as json.
    """

    def __init__(self, project, scope='private', GOTOPATH=None):
        """ Loads json from jfile """
        if GOTOPATH is None:
            GOTOPATH = settings.GOTOPATH
        self.gotopath = GOTOPATH

        self.project = project

        # Creates folder only if not exisiting
        utils.create_project_folder(project, scope, GOTOPATH)

        self.jfile = os.path.join(
            GOTOPATH, 'projects', project, scope, "magicwords.json")

        self.magic = load_magic(self.jfile)

    def reload(self):
        """Reload the magic."""
        self.__init__(self.jfile)

    def _magic_set(self, key, value):
        '''
            Single point of entry for writing magicwords and uris
            to the internal dict.

            This method makes sure that it is encoded
            as utf-8 unicode strings on python2.
        '''
        if sys.version_info[0] == 2:
            if not isinstance(key, unicode):
                key = unicode(key, encoding='utf-8')

            if not isinstance(value, unicode):
                value = unicode(value, encoding='utf-8')

        self.magic[key] = value
        return None

    def save(self):
        """ Saves the magic to jsonfile jfile """
        err = save_magic(self.jfile, self.magic)
        return err

    def add_shortcut(self, magicword, uri):
        """ Adds a magic shortcut if it does not exist yet.
            If it exists, it warns the user.
        """
        if magicword in self.magic.keys():
            return GotoWarning("adding_existing_magicword", 
                magicword=magicword,
                uri=self.magic[magicword],
                newuri=uri
            )

        # setting the magic
        uri = parse_uri(uri)
        err = self._magic_set(magicword, uri)
        if err:
            return err
        return None

    def update_shortcut(self, magicword, uri):
        """ Simply overwrites the content of the magicword """
        uri = parse_uri(uri)
        if magicword in self.magic.keys():
            err = self._magic_set(magicword, uri)
            if err:
                return err
            return None
        else:
            return GotoWarning("updating_nonexisting_magicword", magicword=magicword)


    def rename_shortcut(self, from_magicword, to_magicword, overwrite=False):
        """ Renaming a shortcut 
            Possible errors:

            from_magicword not existing -- checked first
            to_magicword existing -- checked second

            But it is allowed to overwrite to_magicword
            if overwrite == True.
        """
        if from_magicword not in self.magic:
            return GotoWarning('magicword_does_not_exist', magicword=from_magicword)  # noqa

        if not overwrite and to_magicword in self.magic:
            uri = self.magic[to_magicword]
            return GotoWarning('adding_existing_magicword_short', magicword=to_magicword, uri=uri)  # noqa


        from_uri = self.magic[from_magicword]
        del self.magic[from_magicword]
        err = self._magic_set(to_magicword, from_uri)
        
        if err:
            return err
        return None

    def remove_shortcut(self, magicword):
        """ Simply removes a shortcut """
        try:
            self.magic.pop(magicword)
            return None
        except KeyError:
            return GotoWarning("removing_nonexisting_magicword", magicword=magicword)

    def get_uri(self, magicword):
        """ returns the uri from the shortcut of name <magicword> """
        if magicword in self.magic.keys():
            return self.magic[magicword]
        else:
            return None

    def list_shortcuts(self, verbose=False):
        """ Lists all magicwords.
            Optionally may list verbosedly all uris too
            if verbose = True.
        """
        return sorted(self.magic.keys())

    def list_projects(self):
        """List all projects."""
        return sorted(list_projects(self.gotopath))

    def __getitem__(self, key):
        """
            Makes it possible to do print(magic[magiword])
        """
        return self.magic[key]

    def __missing__(self, key):
        """
            ux when magicword is missing
        """
        pass

    def __setitem__(self, key, value):
        """ sets a magicword. Brutal style """
        self._magic_set(key, value)

    def __delitem__(self, key):
        """ removes a magicword. Brutal style """
        self.magic.pop(key)

    def __len__(self):
        """ returns number of magicwords. """
        return len(self.magic.keys())

    def keys(self):
        return self.magic.keys()


def parse_uri(raw_uri):
    ''' Main goal right now: distinguish filesystem paths
        and urls/uris.
        If no scheme is present, assume it is
        a filesystem path. And test for path existence.
        This embeds a rule that all web-urls must start with
        either http:// or https:// for goto to handle them
        properly. One might ponder if goto should magically
        understand that www.gotomagic.com (without scheme)
        is a http-uri.
        TODO: here one could test if the raw_uri
              could work as a valid http(s):// uri,
              in the case where a uri is entered
              with no scheme.
        To handle cases such as: `goto add .`
        if the raw_uri is a path, get the absolute
        path and store that.
    '''
    candidate = os.path.abspath(raw_uri)
    if os.path.exists(candidate):
        return candidate
    else:
        # it is not a file or directory
        # So it must be an uri.
        # But which kind?
        return raw_uri



def list_projects(gotopath):
    """Given a gotopath, lists and filters all project folders as a list of string."""
    path = os.path.join(gotopath, 'projects')
    projects = list(os.walk(path))[0][1]
    return projects


def load_magic(jfile):
    "Loads the magic shortcuts from a json file object"
    if os.path.isfile(jfile) and os.path.getsize(jfile) > 0:
        with codecs.open(jfile, 'r', encoding="utf-8") as f:
            if sys.version_info[0] == 2:
                magic = json.loads(f.read(), encoding='utf-8')
            else:
                magic = json.loads( f.read())
    else:
        magic = {}
    return magic


def save_magic(jfile, magic):
    "saves the magic shortcuts as json into a file"
    with codecs.open(jfile, 'w+', encoding='utf-8') as f:
        try:
            if sys.version_info[0] == 2:
                data = json.dumps(magic, sort_keys=True, indent=4, ensure_ascii=False, encoding='utf-8')
                f.write(unicode(data))
            else:
                data = json.dumps(magic, sort_keys=True, indent=4, ensure_ascii=False)
                f.write(data)
            return None
        except Exception as e:
            return GotoError("magic_could_not_be_saved", message=str(e))



from unittest import TestCase
from string import Formatter

from ..gotomagic.text import GotoWarning, GotoError, error, warning


def get_dict_with_format_args(message):
    """
        Returns a dict of all expected keyword arguments
        in a formated string. Where the value for each key is the key.

        'Hello {name}' would return: {'name': 'name'}
    """
    args = [arg[1] for arg in Formatter().parse(message) if arg[1] is not None]
    return {key: key for key in args}


class TestMagicText(TestCase):

    def test_getting_warning_message(self):
        """
        Testing GotoWarning to return properly formated
        text for all warning messages.
        """

        for message_name in warning.messages.keys():
            raw_message = warning.messages[message_name]
            kwargs = get_dict_with_format_args(raw_message)

            warn = GotoWarning(message_name, **kwargs)
            expected_message = raw_message.format(**kwargs)

            self.assertEqual(warn.message,
                             expected_message,
                             "Message differs for {}".format(message_name))

    def test_getting_error_message(self):
        """
        Testing GotoError to return properly formated
        text for all error messages.
        """

        for message_name in error.messages.keys():
            raw_message = error.messages[message_name]
            kwargs = get_dict_with_format_args(raw_message)

            err = GotoError(message_name, **kwargs)
            expected_message = raw_message.format(**kwargs)

            self.assertEqual(err.message,
                             expected_message,
                             "Message differs for {}".format(message_name))



# coding: utf-8
from __future__ import unicode_literals

from ..gotomagic.text import GotoError, GotoWarning
from ..gotomagic.utils import detect_unmigrated_data


def help():
    return "{0:40}{1}".format('--check-migrate', 'Check if you need to migrate')


def names():
    return ['--check-migrate']


def run(magic, command, args, options):
    ''' checks if we need to migrate underlaying data structure in goto'''

    if detect_unmigrated_data():
            return None, GotoWarning('data_not_migrated')
    else:
        return "Nothing to migrate", None



# git.py
import os
import sys
import git
from .magic import GotoMagic
from . import utils


class GitMagic():
    """ Goto's Black Git box.

        Nah, not really, it is just the place where
        magic shortcuts are created, updated, removed and
        used.

        .. except also commited, and pushed.

        tl-dr: saving magic as json - then gitify the experience.
        https://gitpython.readthedocs.io/en/stable/tutorial.html
    """

    def __init__(self, project):
        """ Loads json from jfile """
        utils.create_project_folder(project, 'shared')

        if not is_git_repo(shared_folder):
            self.repo = git.Repo.init(shared_folder)
        else:
            self.repo = git.Repo(shared_folder)

        shared_magic = GotoMagic(shared_file)
        self.magic = shared_magic

    def reload(self):
        """ reload the underlaying magic """
        self.magic.reload()

    def share(self):
        """
        Interpreting a share as to push to a remote.
        For now.
        """
        gotohub = self.repo.remotes.gotohub
        # may raise git.exc.GitCommandError
        gotohub.pull('master')
        gotohub.push('master')

    def save(self):
        """ Saves the magic to jsonfile jfile """
        self.magic.save()
        self.repo.git.add(self.magic.jfile)
        self.repo.index.commit(self.message)

    def add_shortcut(self, magicword, uri):
        """ Adds a magic shortcut if it does not exist yet.
            If it exists, it warns the user.
        """
        self.magic.add_shortcut(magicword, uri)
        self.message = "add magicword {}".format(magicword)

    def update_shortcut(self, magicword, uri):
        """ Simply overwrites the content of the magicword """
        self.magic.update_shortcut(magicword, uri)
        self.message = "update magicword {}".format(magicword)

    def remove_shortcut(self, magicword):
        """ Simply removes a shortcut """
        self.magic.remove_shortcut(magicword)
        self.message = "rm magicword {}".format(magicword)

    def show_shortcut(self, magicword):
        """ shows a shortcut """
        self.magic.show_shortcut(magicword)

    def get_uri(self, magicword):
        return self.magic.get_uri(magicword)

    def list_shortcuts(self, verbose=False):
        """ Lists all magicwords.
            Optionally may list verbosedly all uris too
            if verbose = True.
        """
        self.magic.list_shortcuts(verbose)

    def __getitem__(self, key):
        """
            Makes it possible to do print(magic[magiword])
        """
        return self.magic[key]

    def __missing__(self, key):
        """
            ux when magicword is missing
        """
        self.magic.__missing__(key)

    def __setitem__(self, key, value):
        """ sets a magicword. Brutal style """
        self.message = "set magicword {}".format(key)
        self.magic[key] = value

    def __delitem__(self, key):
        """ removes a magicword. Brutal style """
        self.message = "rm magicword {}".format(key)
        self.magic.__delitem__(key)

    def __len__(self):
        """ returns number of magicwords. """
        return len(self.magic)

    def keys(self):
        return self.magic.keys()



def is_git_repo(path):
    try:
        _ = git.Repo(path).git_dir
        return True
    except git.exc.InvalidGitRepositoryError:
        return False



# coding: utf-8
from __future__ import unicode_literals
from ..gotomagic.text import GotoError, GotoWarning

def help():
    return "{0:10}{1:30}{2}".format('show', '<magicword>', 'Show url of shortcut')


def names():
    return ['show','--show']


def run(magic, command, args, options):
    """
    Show magicword.
    """

    if (len(args) == 0):
        return None, GotoWarning("missing_magicword", command='show')

    magicword = args[0]
    uri = magic.get_uri(magicword)

    if uri:
        return uri, None
    else:
        return None, GotoWarning("magicword_does_not_exist", magicword=magicword)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
"""
    Text used by GOTO to do UX.

"""
from . import warning, error


class GotoText:
    def __str__(self):
        return self.message

    def __unicode__(self):
        return unicode(self.message, 'utf-8')  # noqa


class GotoError(GotoText):
    def __init__(self, name, **kwargs):
        self.name = name
        self.kwargs = kwargs

    @property
    def message(self):
        # Making retrieval fail fast, if not present, stop and show the error.
        # the missing message should then be made
        text = error.messages.get(self.name, None)
        if text is None:
            raise Exception(
                "Missing error message with name: {}".format(self.name))
        return text.format(**self.kwargs)


class GotoWarning(GotoText):
    def __init__(self, name, **kwargs):
        self.name = name
        self.kwargs = kwargs

    @property
    def message(self):
        # Making retrieval fail fast, if not present, stop and show the error.
        # the missing message should then be made
        text = warning.messages.get(self.name, None)
        if text is None:
            raise Exception(
                "Missing warning message with name: {}".format(self.name))
        return text.format(**self.kwargs)


def print_text(text, **kwargs):
    print(text.format(**kwargs))



# coding: utf-8
from __future__ import unicode_literals

import subprocess
from ..gotomagic.text import GotoError, GotoWarning


def help():
    return "{0:40}{1}".format('subl',  'Opens Sublime Text in code folder')  # noqa


def names():
    return ['subl', 'sublime', '--subl', '--sublime']


def run(magic, command, args, options):
    """
    Launch Sublime Text in the code folder
    """
    magicword = 'code' if len(args) == 0 else args[0]

    uri = magic.get_uri(magicword)

    if not uri:
        return None, GotoWarning("magicword_does_not_exist", magicword=magicword)

    try:
        subprocess.check_call('subl "%s"' % uri, shell=True)
    except subprocess.CalledProcessGotoError:
        return None, GotoError("subl_launch_failed")

    return None, None



# coding: utf-8
from __future__ import absolute_import, unicode_literals
from builtins import dict, str  # redefine dict and str to be py3-like in py2.


# warnings

messages = dict(
    adding_existing_magicword="""
Ah hoy!

    - You already have a magic word named {magicword},
    it points to: {uri}

    If you want to override it use:
        goto update {magicword} {newuri}
    """,

    adding_existing_magicword_short="""
Ah hoy!

    - You already have a magic word named {magicword},
    it points to: {uri}
    """,

    removing_nonexisting_magicword="""
Ah hoy!

    - Attempt to remove non-existing shortcut with name {magicword}

    Are you sure it exists in the project you are in now?
    Type:

        goto list

    to see all shortcuts in this project.
    """,

    updating_nonexisting_magicword="""
Ah hoy!

    - Attempt to update non-existing shortcut with name {magicword}

    Are you sure it exists in the project you are in now?
    Type:

        goto list

    to see all shortcuts in this project.
    """,

    magicword_does_not_exist="""
Ah hoy!

    - The Magic word {magicword} does not exist.

    Are you sure it exists in the project you are in now?
    Type:

        goto list

    to see all shortcuts in this project.
    """,

    no_magicword_named_code="""
Ah hoy!
    - No magicword named code.
    Create it with
        goto add code <path to folder>
    in order to be able to run goto subl.
    """,

    missing_uri="""
Ah hoy!
    - Remember, a shortcut has a name and a target uri (or path).

    Try again by adding another argument with where the shortcut
    should go to:

                goto {command} {magicword} <uri>
    """,

    missing_magicword="""
Ah hoy!
    Error: missing magic word.

    Try again by adding the magicword you want to {command}:

                goto {command} <magicword>
    """,

    missing_both_magicwords="""
Ah hoy!
    Error: missing both magicwords

    Try again by adding the magicword you want to {command}
    and the magicword you want to {command} to:

                goto {command} <from> <to>
    """,

    missing_to_magicword="""
Ah hoy!
    Error: missing second magic word.

    Try again by adding the magicword you want to {command} to:

                goto {command} {magicword} <to>
    """,

    missing_magicword_and_uri="""
Ah hoy!
    - Remember, a shortcut has a name and a target uri (or path).

    Try again by adding both a name and a uri/path to where the shortcut
    should go to:

                goto {command} <magicword> <uri>

    example:
        goto {command} website http://example.org
    """,

    unescaped_ampersand_url_detected="""
Ah hoy!
    - Detected an ampersand (&) in the uri.

    Since goto is ran in a shell, unescaped &'s will make
    goto run in the background, with only half of the uri
    you intended to give it.
    This is expected behaviour in bash/zsh.

    Effectively, chopping your url in half.

    Example: http://example.com?query=param&query2=param
    would loose everything after the & character.

    Try again, but this time wrap the url with "":

        goto {command} {magicword} "http://your-url?a=1&b=2"
    """,

    data_not_migrated="""
Ah hoy!
    - Goto detects that your magicwords needs to be migrated.


    This is because earlier versions of goto stored the
    project data in a different way.
    details: https://github.com/technocake/goto/issues/108
    But no worries, it can be done automagically:

    run the command:
        goto --migrate
    """,

    goto_wont_work_without_migrating_data="""
Ah hoy!

    Goto will not work before data is migrated.
    (the next time you run goto you will be prompted again)
    """,

    not_all_projects_migrated="""
Ah hoy!

    Not all projects could be migrated.
    Please check the output above and move files manually.

    Unmigrated projects: {projects}
""",

    rider_launch_failed="""
Error:  could not launch Rider.

    This is most likely due to not having set up the Rider
    Command Line Launching yet.

    To be able to launch Rider from the command line, see:

        https://www.jetbrains.com/help/rider/Opening_Files_from_Command_Line.html
    """,
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['goto',
 'goto.commands',
 'goto.gotomagic',
 'goto.gotomagic.text',
 'goto.plugins',
 'goto.tests']

package_data = \
{'': ['*']}

install_requires = \
['GitPython>=3.1.14,<4.0.0',
 'future>=0.18.2,<0.19.0',
 'pyperclip>=1.8.2,<2.0.0']

setup_kwargs = {
    'name': 'magicgoto',
    'version': '1.6.1',
    'description': 'Goto - Goto is a magic tool that takes you were you want to be, now.',
    'long_description': None,
    'author': 'technocake',
    'author_email': 'robin.garen@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



name='commands'

from . import help
from . import default
from .commands import commands_list

commands_list = [help, default] + commands_list

commands = {}

for command in commands_list:
    for name in command.names():
        commands[name] = command


def usage():
    '''
        Making this a procedure to avoid  that the help_text could
        be outdated if any later methods inject plugins or other commands.

        This will evaluate the help text for all
        installed commands.
    '''
    help_text, _ = help.run(None, None, None, None)
    return help_text


__all__ = ['usage', 'commands', 'commands_list']



# coding: utf-8
from __future__ import unicode_literals
import pyperclip
from ..gotomagic.text import GotoError, GotoWarning


def help():
    return "{0:10}{1:30}{2}".format('copy', '<magicword>', 'Copy magicword to the clipboard')


def names():
    return ['copy','--copy']


def run(magic, command, args, options):
    """
    Copy uri to clipboard
    """

    if (len(args) == 0):
        return None, GotoWarning("missing_magicword", command='copy')

    word = args[0]
    uri = str(magic.get_uri(word))

    if uri:
        pyperclip.copy(uri)
        return "Copied uri to clipboard", None
    else:
        return None, GotoWarning("magicword_does_not_exist", magicword=word)



# coding: utf-8
from __future__ import unicode_literals
from ..gotomagic.magic import GotoMagic


def help():
    return "{0:10}{1:30}{2}".format('list', '[-v]', 'List all shortcuts')


def names():
    return ['--search', '-s']


def run(magic, command, args, options):
    """
    Search magicwords!
    """
    verbose = '-v' in options or '--verbose' in options
    global_search = '-g' in options or '--global' in options

    if (len(args) == 0):
        return None, GotoWarning("missing_magicword", command='search')

    query = args[0]
    index = all_project_magic(magic, global_search)

    # TODO update api in all_project_magic
    if not index:
        return None, None

    # Search
    magicwords = filter(search(query), index)

    output = "\n".join(magicwords)

    return output, None


def search(query):
    # search magicword
    return lambda shortcut: query in shortcut


def all_project_magic(magic, global_search=False):
    """Loads all projects GotoMagic for searching."""

    if global_search:
        projects = magic.list_projects()
    else:
        projects = [magic.project]

    #TODO handle None

    magic = {}
    index = []

    for project in projects:
        m = GotoMagic(project)
        magic[project] = m
        magicwords = m.list_shortcuts()

        for magicword in magicwords:
            shortformat = "{}.{}".format(project, magicword)
            index.append(shortformat)
            #index.append((project, magicword, shortformat))

    return index


import webbrowser
from ..gotomagic.utils import is_file
from ..gotomagic.text import GotoWarning, GotoError
from ..plugins import open


def help():
    return "<magicword>"


def names():
    return [None, '--default']


def run(magic, command, args, options):
    """
    Default behaviour when no commands are found in the first argument
    """
    verbose = '--verbose' in options or '-v' in options

    output = ""
    for magicword in args:
        uri = magic.get_uri(magicword)

        if uri is None:
            return None, GotoWarning('magicword_does_not_exist', magicword=magicword)  # noqa

        if is_file(uri):
            _output, err = open.run(magic, None, [magicword], verbose)
            if err:
                return None, err
            output += "%s\n" % _output
        else:
            try:
                webbrowser.open_new_tab(uri)
                output += "%s\n" % uri
            except webbrowser.Error:
                return None, GotoError('open_browser_tab_error')

    return output if verbose else None, None
